
/* This file has been autogenerated from source/nptl_cond.c.m4 */
/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#include "marcel.h"
#include "marcel_pmarcel.h"
#include <errno.h>


DEF_MARCEL_PMARCEL(int, condattr_init, (marcel_condattr_t * attr), (attr),
{
	MARCEL_LOG_IN();
	memset(attr, '\0', sizeof(*attr));
	MARCEL_LOG_RETURN(0);
})

DEF_MARCEL_PMARCEL(int, condattr_destroy, (marcel_condattr_t * attr TBX_UNUSED), (attr),
{
	MARCEL_LOG_IN();
	MARCEL_LOG_RETURN(0);
})

DEF_MARCEL_PMARCEL(int, cond_init, (marcel_cond_t * __restrict cond,
				    const marcel_condattr_t * __restrict attr TBX_UNUSED),
		   (cond, attr),
{
	MARCEL_LOG_IN();
 	
	__marcel_init_lock(&cond->__data.__lock);
	cond->__data.__waiting = NULL;

	MARCEL_LOG_RETURN(0);
})

DEF_MARCEL_PMARCEL(int, cond_destroy, (marcel_cond_t * cond), (cond),
{
	MARCEL_LOG_IN();

	if (cond->__data.__waiting != NULL)
		MARCEL_LOG_RETURN(EBUSY);
	__marcel_destroy_lock(&cond->__data.__lock);
	
	MARCEL_LOG_RETURN(0);
})

DEF_MARCEL_PMARCEL(int, cond_signal, (marcel_cond_t * cond), (cond),
{
	MARCEL_LOG_IN();
	__marcel_unlock(&cond->__data.__lock);
	MARCEL_LOG_RETURN(0);
})

DEF_MARCEL_PMARCEL(int, cond_broadcast, (marcel_cond_t * cond), (cond),
{
	MARCEL_LOG_IN();
	
	ma_fastlock_acquire(&cond->__data.__lock);
	while (__marcel_unlock_spinlocked(&cond->__data.__lock));
	ma_fastlock_release(&cond->__data.__lock);

	MARCEL_LOG_RETURN(0);
})

DEF_MARCEL(int, cond_wait, (marcel_cond_t * __restrict cond, marcel_mutex_t * __restrict mutex),
	   (cond, mutex),
{
	MARCEL_LOG_IN();

	/** release mutex and wait on cond */
	ma_fastlock_acquire(&mutex->__data.__lock);
	ma_fastlock_acquire(&cond->__data.__lock);
	__marcel_unlock_spinlocked(&mutex->__data.__lock);
	ma_fastlock_release(&mutex->__data.__lock);
	__marcel_lock_wait(&cond->__data.__lock, ma_self(), 0);
	ma_fastlock_release(&cond->__data.__lock);

	/** reacquire mutex */
	marcel_mutex_lock(mutex);

	MARCEL_LOG_RETURN(0);
})

DEF_PMARCEL(int, cond_wait, (pmarcel_cond_t * __restrict cond, pmarcel_mutex_t * __restrict mutex),
	    (cond, mutex),
{
	MARCEL_LOG_IN();

	/** release the mutex and wait on cond */
	ma_fastlock_acquire(&mutex->__data.__lock);
	ma_fastlock_acquire(&cond->__data.__lock);
	mutex->__data.__owner = 0;
	__marcel_unlock_spinlocked(&mutex->__data.__lock);
	ma_fastlock_release(&mutex->__data.__lock);
	__marcel_lock_wait(&cond->__data.__lock, ma_self(), MA_CHECK_CANCEL);
	ma_fastlock_release(&cond->__data.__lock);

	/** reacquire the mutex and check is I was canceled */
	pmarcel_mutex_lock(mutex);
#ifdef MARCEL_DEVIATION_ENABLED
	if (SELF_GETMEM(canceled) == MARCEL_IS_CANCELED)
		marcel_exit(PMARCEL_CANCELED);
#endif
	MARCEL_LOG_RETURN(0);
})

DEF_MARCEL(int, cond_timedwait, (marcel_cond_t * __restrict cond, marcel_mutex_t * __restrict mutex,
				 const struct timespec *__restrict abstime),
	   (cond, mutex, abstime),
{
	struct timeval now;;
	struct timeval tv;
	unsigned long timeout;
	int ret = 0;
	MARCEL_LOG_IN();

	tv.tv_sec = abstime->tv_sec;
	tv.tv_usec = abstime->tv_nsec / 1000;
	gettimeofday(&now, NULL);
	if (tbx_unlikely(timercmp(&tv, &now, <=)))
		MARCEL_LOG_RETURN(ETIMEDOUT);
	timeout = MA_TIMEVAL_TO_USEC(&tv) - MA_TIMEVAL_TO_USEC(&now);

	/** release mutex and wait on cond */
	ma_fastlock_acquire(&mutex->__data.__lock);
	ma_fastlock_acquire(&cond->__data.__lock);
	__marcel_unlock_spinlocked(&mutex->__data.__lock);
	ma_fastlock_release(&mutex->__data.__lock);
	ret = __marcel_lock_timed_wait(&cond->__data.__lock, ma_self(), timeout, 0);
	ma_fastlock_release(&cond->__data.__lock);

	/** reacquire mutex */
	marcel_mutex_lock(mutex);

	MARCEL_LOG_RETURN(ret);
})

DEF_PMARCEL(int, cond_timedwait, (pmarcel_cond_t * __restrict cond, pmarcel_mutex_t * __restrict mutex,
				  const struct timespec *__restrict abstime),
	    (cond, mutex, abstime),
{
	struct timeval now;
	struct timeval tv;
	unsigned long timeout;
	int ret;

	MARCEL_LOG_IN();

	tv.tv_sec = abstime->tv_sec;
	tv.tv_usec = abstime->tv_nsec / 1000;
	gettimeofday(&now, NULL);
	if (timercmp(&tv, &now, <=))
		MARCEL_LOG_RETURN(ETIMEDOUT);
	timeout = MA_TIMEVAL_TO_USEC(&tv) - MA_TIMEVAL_TO_USEC(&now);

	/** release the mutex and wait on cond */
	ma_fastlock_acquire(&mutex->__data.__lock);
	ma_fastlock_acquire(&cond->__data.__lock);
	mutex->__data.__owner = 0;
	__marcel_unlock_spinlocked(&mutex->__data.__lock);
	ma_fastlock_release(&mutex->__data.__lock);
	ret = __marcel_lock_timed_wait(&cond->__data.__lock, ma_self(), timeout, MA_CHECK_CANCEL);
	ma_fastlock_release(&cond->__data.__lock);

	/** reacquire the mutex and check is I was canceled */
	pmarcel_mutex_lock(mutex);
#ifdef MARCEL_DEVIATION_ENABLED
	if (SELF_GETMEM(canceled) == MARCEL_IS_CANCELED)
		marcel_exit(PMARCEL_CANCELED);
#endif
	MARCEL_LOG_RETURN(ret);
})

DEF_PMARCEL(int, condattr_getpshared, (const pmarcel_condattr_t * __restrict attr, int *__restrict pshared),
	    (attr, pshared),
{
	MARCEL_LOG_IN();
	*pshared = attr->__data.value & 1;
	MARCEL_LOG_RETURN(0);
})

DEF_PMARCEL(int, condattr_setpshared, (pmarcel_condattr_t * attr, int pshared),
	    (attr, pshared),
{
	int *valuep;

	MARCEL_LOG_IN();

	if (tbx_unlikely(pshared != PMARCEL_PROCESS_PRIVATE) && 
	    pshared != PMARCEL_PROCESS_SHARED)
		MARCEL_LOG_RETURN(EINVAL);

	/* For now it is not possible to share a mutex variable.  */
	if (tbx_unlikely(pshared != PMARCEL_PROCESS_PRIVATE))
		MARCEL_LOG_RETURN(ENOTSUP);

	valuep = &(attr)->__data.value;
	*valuep = (*valuep & ~1) | (pshared != PMARCEL_PROCESS_PRIVATE);
	MARCEL_LOG_RETURN(0);
})
