/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#ifdef MARCEL_LIBPTHREAD


#include <unistd.h>
#include <sys/syscall.h>
#include "marcel.h"
#include "marcel_pmarcel.h"


/********************************************************/
#define cancellable_call_generic(ret, name, invocation, ver, proto, ...) \
extern ret LPT_NAME(name) proto ; \
ret LPT_NAME(name) proto { \
	if (tbx_unlikely(!marcel_createdthreads())) { \
		return invocation; \
	} else { \
		int old = __pmarcel_enable_asynccancel(); \
		ret res; \
		res = invocation; \
		__pmarcel_disable_asynccancel(old); \
		return res; \
	} \
} \
versioned_symbol(libpthread, LPT_NAME(name), LIBC_NAME(name), ver); \
DEF___LIBC(ret, name, proto, (args))

#define cancellable_call_ext(ret, name, sysnr, ver, proto, ...)					\
  cancellable_call_generic(ret, name, syscall(sysnr, ##__VA_ARGS__), ver, proto, ##__VA_ARGS__)

#define cancellable_call(ret, name, proto, ...) \
	cancellable_call_ext(ret, name, SYS_##name, GLIBC_2_0, proto, ##__VA_ARGS__)

/*********************lseek***************************/
cancellable_call(off_t, lseek, (int fd, off_t offset, int whence), fd, offset, whence)
#ifdef LINUX_SYS
#  if MA_BITS_PER_LONG < 64
extern off64_t __lseek64(int, off64_t, int);
off64_t __lseek64(int fd, off64_t offset, int whence)
{
	off64_t res;
	int ret;
	if (tbx_unlikely(!marcel_createdthreads())) {
		ret =
		    syscall(SYS__llseek, fd, (off_t) (offset >> 32),
			    (off_t) (offset & 0xffffffff), &res, whence);
	} else {
		int old = __pmarcel_enable_asynccancel();
		ret =
		    syscall(SYS__llseek, fd, (off_t) (offset >> 32),
			    (off_t) (offset & 0xffffffff), &res, whence);
		__pmarcel_disable_asynccancel(old);
	}
	return ret ? ret : res;
}

DEF_STRONG_ALIAS(__lseek64, lseek64)

extern loff_t __llseek(int, loff_t, int);
loff_t __llseek(int fd, loff_t offset, int whence)
{
	loff_t res;
	int ret;
	if (tbx_unlikely(!marcel_createdthreads())) {
		ret =
		    syscall(SYS__llseek, fd, (off_t) (offset >> 32),
			    (off_t) (offset & 0xffffffff), &res, whence);
	} else {
		int old = __pmarcel_enable_asynccancel();
		ret =
		    syscall(SYS__llseek, fd, (off_t) (offset >> 32),
			    (off_t) (offset & 0xffffffff), &res, whence);
		__pmarcel_disable_asynccancel(old);
	}
	return ret ? ret : res;
}

DEF_STRONG_ALIAS(__llseek, llseek)
#  else
DEF_WEAK_ALIAS(LPT_NAME(lseek), lseek64)
DEF_WEAK_ALIAS(LPT_NAME(lseek), llseek)
#  endif
#endif /** LINUX_SYS **/

cancellable_call(int, close, (int fd), (fd))
cancellable_call(int, fcntl, (int fd, int cmd, long arg), fd, cmd, arg)
cancellable_call(int, fsync, (int fd), (fd))
cancellable_call(int, open, (const char *path, int flags, mode_t mode), path, flags, mode)
cancellable_call_ext(int, open64, SYS_open, GLIBC_2_2,
		     (const char *path, int flags, mode_t mode), path, flags, mode)
#ifdef SYS_pread
cancellable_call_ext(ssize_t, pread, SYS_pread, GLIBC_2_2,
		     (int fd, void *buf, size_t count, off_t pos), fd, buf, count, pos)
#endif
#ifdef SYS_pread64
cancellable_call_ext(ssize_t, pread64, SYS_pread64, GLIBC_2_2,
		     (int fd, void *buf, size_t count, off64_t pos), fd, buf, count, pos)
#  ifndef SYS_pread
cancellable_call_ext(ssize_t, pread, SYS_pread64, GLIBC_2_2,
		     (int fd, void *buf, size_t count, off_t pos), fd, buf, count, pos)
#  endif
#endif
#ifdef SYS_pwrite
cancellable_call_ext(ssize_t, pwrite, SYS_pwrite, GLIBC_2_2,
		     (int fd, const void *buf, size_t count, off_t pos), fd, buf, count, pos)
#endif
#ifdef SYS_pwrite64
cancellable_call_ext(ssize_t, pwrite64, SYS_pwrite64, GLIBC_2_2,
		     (int fd, const void *buf, size_t count, off64_t pos), fd, buf, count, pos)
#  ifndef SYS_pwrite
cancellable_call_ext(ssize_t, pwrite, SYS_pwrite64, GLIBC_2_2,
		     (int fd, const void *buf, size_t count, off64_t pos), fd, buf, count, pos)
#  endif
#endif
cancellable_call(ssize_t, read, (int fd, void *buf, size_t count), fd, buf, count)
cancellable_call(ssize_t, write, (int fd, const void *buf, size_t count), fd, buf, count)
#ifdef SYS_connect
cancellable_call(int, connect, (int sockfd, const struct sockaddr * serv_addr,
				socklen_t addrlen), sockfd, serv_addr, addrlen)
#endif
#ifdef SYS_accept
cancellable_call(int, accept, (int sockfd, struct sockaddr * addr,
			       socklen_t * addrlen), sockfd, addr, addrlen)
#endif
#ifdef SYS_sendto
cancellable_call(ssize_t, sendto,
		 (int s, const void *buf, size_t len, int flags,
		  const struct sockaddr * to, socklen_t tolen), s, buf, len, flags, to,
		 tolen)
#endif
#ifdef SYS_recvfrom
cancellable_call(ssize_t, recvfrom,
		 (int s, void *buf, size_t len, int flags, struct sockaddr * from,
		  socklen_t * fromlen), s, buf, len, flags, from, fromlen)
#endif
ssize_t send(int s, const void *buf, size_t len, int flags)
{
	return sendto(s, buf, len, flags, NULL, 0);
}

ssize_t recv(int s, void *buf, size_t len, int flags)
{
	return recvfrom(s, buf, len, flags, NULL, NULL);
}


#endif /** MARCEL_LIBPTHREAD **/
